<?php declare(strict_types=1);

namespace Infomaker\Everyware\Twig;

use Twig_Error;
use Twig_SimpleFilter;
use Twig_SimpleFunction;


abstract class ViewSetup
{
    protected $cache = false;
    protected $cacheTime = 86400;
    protected $debug = true;

    /**
     * Contains Wordpress functions that will be added to the list off functions
     * that will be available in the templates
     * function: [func_name] => [custom_name]
     *
     * @var array
     */
    protected $wpFunctions = [];

    /**
     * Contains functions that will be available in the templates.
     * If the same function name is added several times only the last one is kept.
     * @var array
     */
    protected static $functions = [];

    /**
     * Contains filters that will be available in the templates.
     * If the same filters name is added several times only the last one is kept.
     * @var array
     */
    protected static $filters = [];
    /**
     * Contains filters that will be available in the templates.
     * @var array
     */
    protected static $globals = [];

    /**
     * Contains the FilesystemLoader aka all template-paths available in the templates.
     * @var FilesystemLoader
     */
    protected static $loader;

    public function __construct()
    {
        //Add global device variable
        if (\defined('EVERY_DEVICE')) {
            $this->addGlobal('device', EVERY_DEVICE);
        }

        // Wordpress functions
        // ======================================================
        foreach ($this->wpFunctions as $wpFunction => $name) {
            $this->addWpFunction($wpFunction, $name);
        }

        $this->init();
    }

    public function registerTwigFolder(string $namespace, string $folderPath, array $exclude = null): self
    {
        if (is_dir($folderPath)) {
            try {
                $this->getLoader()->addPathAndSubfolders($folderPath, $namespace, $exclude ?? []);
            } catch (Twig_Error $e) {
                new \Exception($e->getMessage());
            }
        }

        return $this;
    }

    protected function getLoader(): FilesystemLoader
    {
        if (static::$loader === null) {
            static::$loader = new FilesystemLoader();
        }

        return static::$loader;
    }

    public function addFilter(string $name, $callable, array $options = null): self
    {
        static::$filters[$name] = new Twig_SimpleFilter($name, $callable, $options ?? []);

        return $this;
    }

    public function addFunction(string $name, $callable, array $options = null): self
    {
        static::$functions[$name] = new Twig_SimpleFunction($name, $callable, $options ?? []);

        return $this;

    }

    public function addGlobal(string $key, $value): self
    {
        static::$globals[$key] = $value;

        return $this;
    }

    public function addWpFunction(string $functionName, string $name = null)
    {
        $name = ( ! empty($name) && $name !== null) ? $name : $functionName;
        $this->addFunction($name, $functionName);

        return $this;
    }

    protected function init(array $options = []): void
    {
        // Use the abstract class in child-classes to customize setup
        $this->setupTwig();
        $this->setLoaderCache();

        $twig = EveryTwig::get_instance()->init([$this->getLoader()], static::$functions, static::$filters,
            array_replace([
                'debug' => $this->debug,
                'cache' => $this->cache,
                'cachetime' => $this->cacheTime,
            ], $options));

        foreach (static::$globals as $key => $value) {
            $twig->addGlobal($key, $value);
        }
    }

    abstract public function setupTwig();

    public static function setup(): ViewSetup
    {
        return new static();
    }

    private function setLoaderCache(): void
    {
        $this->getLoader()->setCacheTime($this->cache ? $this->cacheTime : 10);
    }
}
